from functools import partial
from sklearn.base import ClassifierMixin, RegressorMixin
from sklearn.utils import all_estimators
from ..multitask import MultitaskClassifier, SimpleMultitaskClassifier
from ..mts import ClassicalMTS


removed_classifiers = [
    "ClassifierChain",
    "ComplementNB",
    "GaussianProcessClassifier",
    "GradientBoostingClassifier",
    "HistGradientBoostingClassifier",
    "MultiOutputClassifier",
    "MultinomialNB",
    "OneVsOneClassifier",
    "OneVsRestClassifier",
    "OutputCodeClassifier",
    "RadiusNeighborsClassifier",
    "StackingClassifier",
    "VotingClassifier",
]

removed_regressors = [
    "TheilSenRegressor",
    "ARDRegression",
    "CCA",
    "GaussianProcessRegressor",
    "GradientBoostingRegressor",
    "HistGradientBoostingRegressor",
    "IsotonicRegression",
    "KernelRidge",
    "MultiOutputRegressor",
    "MultiTaskElasticNet",
    "MultiTaskElasticNetCV",
    "MultiTaskLasso",
    "MultiTaskLassoCV",
    "NuSVR",
    "OrthogonalMatchingPursuit",
    "OrthogonalMatchingPursuitCV",
    "PLSCanonical",
    "PLSRegression",
    "RadiusNeighborsRegressor",
    "RegressorChain",
    "StackingRegressor",
    "SVR",
    "VotingRegressor",
]

CLASSIFIERS = [
    ("CustomClassifier(" + est[0] + ")", est[1])
    for est in all_estimators()
    if (
        issubclass(est[1], ClassifierMixin)
        and (est[0] not in removed_classifiers)
    )
]

DEEPCLASSIFIERS = [
    ("DeepCustomClassifier(" + est[0] + ")", est[1])
    for est in all_estimators()
    if (
        issubclass(est[1], ClassifierMixin)
        and (est[0] not in removed_classifiers)
    )
]

MULTITASKCLASSIFIERS = [
    (
        "MultitaskClassifier(" + est[0] + ")",
        partial(MultitaskClassifier, obj=est[1]()),
    )
    for est in all_estimators()
    if (
        issubclass(est[1], RegressorMixin)
        and (est[0] not in removed_regressors)
    )
]

DEEPMULTITASKCLASSIFIERS = [
    (
        "DeepMultitaskClassifier(" + est[0] + ")",
        partial(MultitaskClassifier, obj=est[1]()),
    )
    for est in all_estimators()
    if (
        issubclass(est[1], RegressorMixin)
        and (est[0] not in removed_regressors)
    )
]

SIMPLEMULTITASKCLASSIFIERS = [
    (
        "SimpleMultitaskClassifier(" + est[0] + ")",
        partial(SimpleMultitaskClassifier, obj=est[1]()),
    )
    for est in all_estimators()
    if (
        issubclass(est[1], RegressorMixin)
        and (est[0] not in removed_regressors)
    )
]

DEEPSIMPLEMULTITASKCLASSIFIERS = [
    (
        "DeepSimpleMultitaskClassifier(" + est[0] + ")",
        partial(SimpleMultitaskClassifier, obj=est[1]()),
    )
    for est in all_estimators()
    if (
        issubclass(est[1], RegressorMixin)
        and (est[0] not in removed_regressors)
    )
]

REGRESSORS = [
    ("CustomRegressor(" + est[0] + ")", est[1])
    for est in all_estimators()
    if (
        issubclass(est[1], RegressorMixin)
        and (est[0] not in removed_regressors)
    )
]

DEEPREGRESSORS = [
    ("DeepCustomRegressor(" + est[0] + ")", est[1])
    for est in all_estimators()
    if (
        issubclass(est[1], RegressorMixin)
        and (est[0] not in removed_regressors)
    )
]

REGRESSORSMTS = [
    ("MTS(" + est[0] + ")", est[1])
    for est in all_estimators()
    if (
        issubclass(est[1], RegressorMixin)
        and (est[0] not in removed_regressors)
    )
]

DEEPREGRESSORSMTS = [
    ("DeepMTS(" + est[0] + ")", est[1])
    for est in all_estimators()
    if (
        issubclass(est[1], RegressorMixin)
        and (est[0] not in removed_regressors)
    )
]

CLASSICALMTS = [
    ("ClassicalMTS(" + est[0] + ")", est[1])
    for est in [
        ("VAR", partial(ClassicalMTS, model="VAR")),
        ("VECM", partial(ClassicalMTS, model="VECM")),
    ]
]



from locale import normalize
import numpy as np
import pickle
from collections import namedtuple
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.model_selection import train_test_split
from sklearn.ensemble import ExtraTreesRegressor
from sklearn.preprocessing import StandardScaler
from scipy.stats import gaussian_kde
from tqdm import tqdm
from ..nonconformist import (
    ClassifierAdapter,
    IcpClassifier,
    TcpClassifier,
    ClassifierNc,
    MarginErrFunc,
)


class PredictionSet(BaseEstimator, ClassifierMixin):
    """Class PredictionSet: Obtain prediction sets.

    Attributes:

        obj: an object;
            fitted object containing methods `fit` and `predict`

        method: a string;
            method for constructing the prediction sets.
            Currently "icp" (default, inductive conformal) and "tcp" (transductive conformal)

        level: a float;
            Confidence level for prediction sets. Default is None,
            95 is equivalent to a miscoverage error of 5 (%)

        seed: an integer;
            Reproducibility of fit (there's a random split between fitting and calibration data)
    """

    def __init__(
        self,
        obj,
        method="icp",
        level=None,
        seed=123,
    ):

        self.obj = obj
        self.method = method
        self.level = level
        self.seed = seed
        if self.level is not None:
            self.alpha_ = 1 - self.level / 100
        self.quantile_ = None
        self.icp_ = None
        self.tcp_ = None

        if self.method == "icp":
            self.icp_ = IcpClassifier(
                ClassifierNc(ClassifierAdapter(self.obj), MarginErrFunc()),
            )
        elif self.method == "tcp":
            self.tcp_ = TcpClassifier(
                ClassifierNc(ClassifierAdapter(self.obj), MarginErrFunc()),
            )
        else:
            raise ValueError("`self.method` must be in ('icp', 'tcp')")

    def fit(self, X, y, sample_weight=None, **kwargs):
        """Fit the `method` to training data (X, y).

        Args:

            X: array-like, shape = [n_samples, n_features];
                Training set vectors, where n_samples is the number
                of samples and n_features is the number of features.

            y: array-like, shape = [n_samples, ]; Target values.

            sample_weight: array-like, shape = [n_samples]
                Sample weights.

        """
        if self.method == "icp":

            X_train, X_calibration, y_train, y_calibration = train_test_split(
                X, y, test_size=0.5, random_state=self.seed
            )
            self.icp_.fit(X_train, y_train)
            self.icp_.calibrate(X_calibration, y_calibration)

        elif self.method == "tcp":

            self.tcp_.fit(X, y)

        return self

    def predict(self, X, **kwargs):
        """Obtain predictions and prediction sets

        Args:

            X: array-like, shape = [n_samples, n_features];
                Testing set vectors, where n_samples is the number
                of samples and n_features is the number of features.

        """

        if self.method == "icp":
            return self.icp_.predict(X, significance=self.alpha_, **kwargs)

        elif self.method == "tcp":
            return self.tcp_.predict(X, significance=self.alpha_, **kwargs)

        else:
            raise ValueError("`self.method` must be in ('icp', 'tcp')")

    def predict_proba(self, X):
        predictions = self.predict(X)
        return np.eye(len(np.unique(predictions)))[predictions]



import numpy as np
import pandas as pd


def getsims(input_data, ix):

    if isinstance(input_data[0], pd.DataFrame):  # kde
        n_sims = len(input_data)
        res = [input_data[i].iloc[:, ix].values for i in range(n_sims)]
        return np.asarray(res).T

    if isinstance(input_data[0], tuple):  # GP posterior
        h = len(input_data)
        n_sims = len(input_data[0][0][0])
        res = [
            [input_data[hx][ix][0][i] for i in range(n_sims)] for hx in range(h)
        ]
        return np.asarray(res)



# Authors: Thierry Moudiki
#
# License: BSD 3

import numpy as np
import sklearn.metrics as skm2
from .custom import Custom
from ..predictionset import PredictionSet
from ..utils import matrixops as mo
from sklearn.base import ClassifierMixin


class CustomClassifier(Custom, ClassifierMixin):
    """Custom Classification model

    Attributes:

        obj: object
            any object containing a method fit (obj.fit()) and a method predict
            (obj.predict())

        n_hidden_features: int
            number of nodes in the hidden layer

        activation_name: str
            activation function: 'relu', 'tanh', 'sigmoid', 'prelu' or 'elu'

        a: float
            hyperparameter for 'prelu' or 'elu' activation function

        nodes_sim: str
            type of simulation for the nodes: 'sobol', 'hammersley', 'halton',
            'uniform'

        bias: boolean
            indicates if the hidden layer contains a bias term (True) or not
            (False)

        dropout: float
            regularization parameter; (random) percentage of nodes dropped out
            of the training

        direct_link: boolean
            indicates if the original predictors are included (True) in model''s
            fitting or not (False)

        n_clusters: int
            number of clusters for 'kmeans' or 'gmm' clustering (could be 0:
                no clustering)

        cluster_encode: bool
            defines how the variable containing clusters is treated (default is one-hot)
            if `False`, then labels are used, without one-hot encoding

        type_clust: str
            type of clustering method: currently k-means ('kmeans') or Gaussian
            Mixture Model ('gmm')

        type_scaling: a tuple of 3 strings
            scaling methods for inputs, hidden layer, and clustering respectively
            (and when relevant).
            Currently available: standardization ('std') or MinMax scaling ('minmax')

        col_sample: float
            percentage of covariates randomly chosen for training

        row_sample: float
            percentage of rows chosen for training, by stratified bootstrapping

        level: float
            confidence level for prediction sets. Default is None.

        pi_method: str
            method for constructing the prediction sets: 'icp', 'tcp' if level is not None. Default is 'icp'.

        seed: int
            reproducibility seed for nodes_sim=='uniform'

        backend: str
            "cpu" or "gpu" or "tpu"

    Examples:

    Note: it's better to use the `DeepClassifier` or `LazyDeepClassifier` classes directly

    ```python
    import nnetsauce as ns
    from sklearn.ensemble import RandomForestClassifier
    from sklearn.model_selection import train_test_split
    from sklearn.datasets import load_digits
    from time import time

    digits = load_digits()
    X = digits.data
    y = digits.target
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2,
                                                        random_state=123)

    # layer 1 (base layer) ----
    layer1_regr = RandomForestClassifier(n_estimators=10, random_state=123)

    start = time()

    layer1_regr.fit(X_train, y_train)

    # Accuracy in layer 1
    print(layer1_regr.score(X_test, y_test))

    # layer 2 using layer 1 ----
    layer2_regr = ns.CustomClassifier(obj = layer1_regr, n_hidden_features=5,
                            direct_link=True, bias=True,
                            nodes_sim='uniform', activation_name='relu',
                            n_clusters=2, seed=123)
    layer2_regr.fit(X_train, y_train)

    # Accuracy in layer 2
    print(layer2_regr.score(X_test, y_test))

    # layer 3 using layer 2 ----
    layer3_regr = ns.CustomClassifier(obj = layer2_regr, n_hidden_features=10,
                            direct_link=True, bias=True, dropout=0.7,
                            nodes_sim='uniform', activation_name='relu',
                            n_clusters=2, seed=123)
    layer3_regr.fit(X_train, y_train)

    # Accuracy in layer 3
    print(layer3_regr.score(X_test, y_test))

    print(f"Elapsed {time() - start}")
    ```

    """

    # construct the object -----

    def __init__(
        self,
        obj,
        n_hidden_features=5,
        activation_name="relu",
        a=0.01,
        nodes_sim="sobol",
        bias=True,
        dropout=0,
        direct_link=True,
        n_clusters=2,
        cluster_encode=True,
        type_clust="kmeans",
        type_scaling=("std", "std", "std"),
        col_sample=1,
        row_sample=1,
        level=None,
        pi_method="icp",
        seed=123,
        backend="cpu",
    ):
        super().__init__(
            obj=obj,
            n_hidden_features=n_hidden_features,
            activation_name=activation_name,
            a=a,
            nodes_sim=nodes_sim,
            bias=bias,
            dropout=dropout,
            direct_link=direct_link,
            n_clusters=n_clusters,
            cluster_encode=cluster_encode,
            type_clust=type_clust,
            type_scaling=type_scaling,
            col_sample=col_sample,
            row_sample=row_sample,
            seed=seed,
            backend=backend,
        )
        self.level = level
        self.pi_method = pi_method
        self.type_fit = "classification"
        if self.level is not None:
            self.obj = PredictionSet(
                self.obj, level=self.level, method=self.pi_method
            )

    def fit(self, X, y, sample_weight=None, **kwargs):
        """Fit custom model to training data (X, y).

        Parameters:

            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number
                of samples and n_features is the number of features.

            y: array-like, shape = [n_samples]
                Target values.

            sample_weight: array-like, shape = [n_samples]
                Sample weights.

            **kwargs: additional parameters to be passed to
                        self.cook_training_set or self.obj.fit

        Returns:

            self: object
        """

        output_y, scaled_Z = self.cook_training_set(y=y, X=X, **kwargs)
        self.n_classes_ = len(np.unique(y))  # for compatibility with sklearn

        if self.level is not None:
            self.obj = PredictionSet(
                obj=self.obj, method=self.pi_method, level=self.level
            )

        # if sample_weights, else: (must use self.row_index)
        if sample_weight is not None:
            self.obj.fit(
                scaled_Z,
                output_y,
                sample_weight=sample_weight[self.index_row_].ravel(),
                # **kwargs
            )

            return self

        # if sample_weight is None:
        self.obj.fit(scaled_Z, output_y)
        self.classes_ = np.unique(y)  # for compatibility with sklearn
        self.n_classes_ = len(self.classes_)  # for compatibility with sklearn

        return self

    def partial_fit(self, X, y, sample_weight=None, **kwargs):
        """Partial fit custom model to training data (X, y).

        Parameters:

            X: {array-like}, shape = [n_samples, n_features]
                Subset of training vectors, where n_samples is the number
                of samples and n_features is the number of features.

            y: array-like, shape = [n_samples]
                Subset of target values.

            sample_weight: array-like, shape = [n_samples]
                Sample weights.

            **kwargs: additional parameters to be passed to
                        self.cook_training_set or self.obj.fit

        Returns:

            self: object
        """

        output_y, scaled_Z = self.cook_training_set(y=y, X=X, **kwargs)
        self.n_classes_ = len(np.unique(y))  # for compatibility with sklearn

        # if sample_weights, else: (must use self.row_index)
        if sample_weight is not None:
            try:
                self.obj.partial_fit(
                    scaled_Z,
                    output_y,
                    sample_weight=sample_weight[self.index_row_].ravel(),
                    # **kwargs
                )
            except:
                NotImplementedError

            return self

        # if sample_weight is None:
        try:
            self.obj.fit(scaled_Z, output_y)
        except:
            raise NotImplementedError

        self.classes_ = np.unique(y)  # for compatibility with sklearn
        self.n_classes_ = len(self.classes_)  # for compatibility with sklearn

        return self

    def predict(self, X, **kwargs):
        """Predict test data X.

        Parameters:

            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number
                of samples and n_features is the number of features.

            **kwargs: additional parameters to be passed to
                    self.cook_test_set

        Returns:

            model predictions: {array-like}
        """

        if len(X.shape) == 1:
            n_features = X.shape[0]
            new_X = mo.rbind(
                X.reshape(1, n_features),
                np.ones(n_features).reshape(1, n_features),
            )

            return (
                self.obj.predict(self.cook_test_set(new_X, **kwargs), **kwargs)
            )[0]

        return self.obj.predict(self.cook_test_set(X, **kwargs), **kwargs)

    def predict_proba(self, X, **kwargs):
        """Predict probabilities for test data X.

        Args:

            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number
                of samples and n_features is the number of features.

            **kwargs: additional parameters to be passed to
                    self.cook_test_set

        Returns:

            probability estimates for test data: {array-like}
        """

        if len(X.shape) == 1:
            n_features = X.shape[0]
            new_X = mo.rbind(
                X.reshape(1, n_features),
                np.ones(n_features).reshape(1, n_features),
            )

            return (
                self.obj.predict_proba(
                    self.cook_test_set(new_X, **kwargs), **kwargs
                )
            )[0]

        return self.obj.predict_proba(self.cook_test_set(X, **kwargs), **kwargs)



import numpy as np
import pyvinecopulib as pv


def vinecopula_sample(x, n_samples=10, method="vine-tll", random_state=123):
    u = pv.to_pseudo_obs(x)
    method_name = method.replace("scp-vine-", "")
    method_name = method_name.replace("scp2-vine-", "")
    method_name = method_name.replace("vine-", "")
    controls = pv.FitControlsVinecop(
        family_set=[getattr(pv.BicopFamily, method_name)]
    )
    cop = pv.Vinecop(u, controls=controls)
    u_sim = cop.simulate(n_samples, seeds=[random_state])
    p = x.shape[1]
    return np.asarray([np.quantile(x[:, i], u_sim[:, i]) for i in range(p)]).T



# Authors: Thierry Moudiki
#
# License: BSD 3 Clear

import copy
import numpy as np
import pandas as pd
import platform
import warnings
import sklearn.metrics as skm

from functools import partial
from sklearn.base import BaseEstimator, ClassifierMixin, RegressorMixin
from ..utils import activations as ac
from ..utils import matrixops as mo
from ..utils import misc as mx
from ..simulation import (
    generate_sobol,
    generate_uniform,
    generate_hammersley,
    generate_halton,
)
from ..sampling import SubSampler


try:
    import jax.nn as jnn
    import jax.numpy as jnp
except ImportError:
    pass


class Base(BaseEstimator):
    """Base model from which all the other classes inherit.

    This class contains the most important data preprocessing/feature engineering methods.

    Parameters:

        n_hidden_features: int
            number of nodes in the hidden layer

        activation_name: str
            activation function: 'relu', 'tanh', 'sigmoid', 'prelu' or 'elu'

        a: float
            hyperparameter for 'prelu' or 'elu' activation function

        nodes_sim: str
            type of simulation for hidden layer nodes: 'sobol', 'hammersley', 'halton',
            'uniform'

        bias: boolean
            indicates if the hidden layer contains a bias term (True) or
            not (False)

        dropout: float
            regularization parameter; (random) percentage of nodes dropped out
            of the training

        direct_link: boolean
            indicates if the original features are included (True) in model's
            fitting or not (False)

        n_clusters: int
            number of clusters for type_clust='kmeans' or type_clust='gmm'
            clustering (could be 0: no clustering)

        cluster_encode: bool
            defines how the variable containing clusters is treated (default is one-hot);
            if `False`, then labels are used, without one-hot encoding

        type_clust: str
            type of clustering method: currently k-means ('kmeans') or Gaussian
            Mixture Model ('gmm')

        type_scaling: a tuple of 3 strings
            scaling methods for inputs, hidden layer, and clustering respectively
            (and when relevant).
            Currently available: standardization ('std') or MinMax scaling ('minmax') or robust scaling ('robust') or  max absolute scaling ('maxabs')

        col_sample: float
            percentage of features randomly chosen for training

        row_sample: float
            percentage of rows chosen for training, by stratified bootstrapping

        seed: int
            reproducibility seed for nodes_sim=='uniform', clustering and dropout

        backend: str
            "cpu" or "gpu" or "tpu"

    """

    # construct the object -----

    def __init__(
        self,
        n_hidden_features=5,
        activation_name="relu",
        a=0.01,
        nodes_sim="sobol",
        bias=True,
        dropout=0,
        direct_link=True,
        n_clusters=2,
        cluster_encode=True,
        type_clust="kmeans",
        type_scaling=("std", "std", "std"),
        col_sample=1,
        row_sample=1,
        seed=123,
        backend="cpu",
    ):
        # input checks -----

        sys_platform = platform.system()

        if (sys_platform == "Windows") and (backend in ("gpu", "tpu")):
            warnings.warn(
                "No GPU/TPU computing on Windows yet, backend set to 'cpu'"
            )
            backend = "cpu"

        assert activation_name in (
            "relu",
            "tanh",
            "sigmoid",
            "prelu",
            "elu",
        ), "'activation_name' must be in ('relu', 'tanh', 'sigmoid','prelu', 'elu')"

        assert nodes_sim in (
            "sobol",
            "hammersley",
            "uniform",
            "halton",
        ), "'nodes_sim' must be in ('sobol', 'hammersley', 'uniform', 'halton')"

        assert type_clust in (
            "kmeans",
            "gmm",
        ), "'type_clust' must be in ('kmeans', 'gmm')"

        assert (len(type_scaling) == 3) & all(
            type_scaling[i] in ("minmax", "std", "robust", "maxabs")
            for i in range(len(type_scaling))
        ), "'type_scaling' must have length 3, and available scaling methods are 'minmax' scaling, standardization ('std'), robust scaling ('robust') and max absolute ('maxabs')"

        assert (col_sample >= 0) & (
            col_sample <= 1
        ), "'col_sample' must be comprised between 0 and 1 (both included)"

        assert backend in (
            "cpu",
            "gpu",
            "tpu",
        ), "must have 'backend' in ('cpu', 'gpu', 'tpu')"

        self.n_hidden_features = n_hidden_features
        self.activation_name = activation_name
        self.a = a
        self.nodes_sim = nodes_sim
        self.bias = bias
        self.seed = seed
        self.backend = backend
        self.dropout = dropout
        self.direct_link = direct_link
        self.cluster_encode = cluster_encode
        self.type_clust = type_clust
        self.type_scaling = type_scaling
        self.col_sample = col_sample
        self.row_sample = row_sample
        self.n_clusters = n_clusters
        if isinstance(self, RegressorMixin):
            self.type_fit = "regression"
        elif isinstance(self, ClassifierMixin):
            self.type_fit = "classification"
        self.subsampler_ = None
        self.index_col_ = None
        self.index_row_ = True
        self.clustering_obj_ = None
        self.clustering_scaler_ = None
        self.nn_scaler_ = None
        self.scaler_ = None
        self.encoder_ = None
        self.W_ = None
        self.X_ = None
        self.y_ = None
        self.y_mean_ = None
        self.beta_ = None

        # activation function -----
        if sys_platform in ("Linux", "Darwin"):
            activation_options = {
                "relu": ac.relu if (self.backend == "cpu") else jnn.relu,
                "tanh": np.tanh if (self.backend == "cpu") else jnp.tanh,
                "sigmoid": (
                    ac.sigmoid if (self.backend == "cpu") else jnn.sigmoid
                ),
                "prelu": partial(ac.prelu, a=a),
                "elu": (
                    partial(ac.elu, a=a)
                    if (self.backend == "cpu")
                    else partial(jnn.elu, a=a)
                ),
            }
        else:  # on Windows currently, no JAX
            activation_options = {
                "relu": (
                    ac.relu if (self.backend == "cpu") else NotImplementedError
                ),
                "tanh": (
                    np.tanh if (self.backend == "cpu") else NotImplementedError
                ),
                "sigmoid": (
                    ac.sigmoid
                    if (self.backend == "cpu")
                    else NotImplementedError
                ),
                "prelu": partial(ac.prelu, a=a),
                "elu": (
                    partial(ac.elu, a=a)
                    if (self.backend == "cpu")
                    else NotImplementedError
                ),
            }
        self.activation_func = activation_options[activation_name]

    # "preprocessing" methods to be inherited -----

    def encode_clusters(self, X=None, predict=False, **kwargs):  #
        """Create new covariates with kmeans or GMM clustering

        Parameters:

            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number
                of samples and n_features is the number of features.

            predict: boolean
                is False on training set and True on test set

            **kwargs:
                additional parameters to be passed to the
                clustering method

        Returns:

            Clusters' matrix, one-hot encoded: {array-like}

        """

        np.random.seed(self.seed)

        if X is None:
            X = self.X_

        if isinstance(X, pd.DataFrame):
            X = copy.deepcopy(X.values.astype(float))

        if predict is False:  # encode training set
            # scale input data before clustering
            self.clustering_scaler_, scaled_X = mo.scale_covariates(
                X, choice=self.type_scaling[2]
            )

            self.clustering_obj_, X_clustered = mo.cluster_covariates(
                scaled_X,
                self.n_clusters,
                self.seed,
                type_clust=self.type_clust,
                **kwargs
            )

            if self.cluster_encode == True:
                return mo.one_hot_encode(X_clustered, self.n_clusters).astype(
                    np.float16
                )

            return X_clustered.astype(np.float16)

        # if predict == True, encode test set
        X_clustered = self.clustering_obj_.predict(
            self.clustering_scaler_.transform(X)
        )

        if self.cluster_encode == True:
            return mo.one_hot_encode(X_clustered, self.n_clusters).astype(
                np.float16
            )

        return X_clustered.astype(np.float16)

    def create_layer(self, scaled_X, W=None):
        """Create hidden layer.

        Parameters:

            scaled_X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number
                of samples and n_features is the number of features

            W: {array-like}, shape = [n_features, hidden_features]
                if provided, constructs the hidden layer with W; otherwise computed internally

        Returns:

            Hidden layer matrix: {array-like}

        """

        n_features = scaled_X.shape[1]

        # hash_sim = {
        #         "sobol": generate_sobol,
        #         "hammersley": generate_hammersley,
        #         "uniform": generate_uniform,
        #         "halton": generate_halton
        #     }

        if self.bias is False:  # no bias term in the hidden layer
            if W is None:
                if self.nodes_sim == "sobol":
                    self.W_ = generate_sobol(
                        n_dims=n_features,
                        n_points=self.n_hidden_features,
                        seed=self.seed,
                    )
                elif self.nodes_sim == "hammersley":
                    self.W_ = generate_hammersley(
                        n_dims=n_features,
                        n_points=self.n_hidden_features,
                        seed=self.seed,
                    )
                elif self.nodes_sim == "uniform":
                    self.W_ = generate_uniform(
                        n_dims=n_features,
                        n_points=self.n_hidden_features,
                        seed=self.seed,
                    )
                else:
                    self.W_ = generate_halton(
                        n_dims=n_features,
                        n_points=self.n_hidden_features,
                        seed=self.seed,
                    )

                # self.W_ = hash_sim[self.nodes_sim](
                #             n_dims=n_features,
                #             n_points=self.n_hidden_features,
                #             seed=self.seed,
                #         )

                assert (
                    scaled_X.shape[1] == self.W_.shape[0]
                ), "check dimensions of covariates X and matrix W"

                return mo.dropout(
                    x=self.activation_func(
                        mo.safe_sparse_dot(
                            a=scaled_X, b=self.W_, backend=self.backend
                        )
                    ),
                    drop_prob=self.dropout,
                    seed=self.seed,
                )

            # W is not none
            assert (
                scaled_X.shape[1] == W.shape[0]
            ), "check dimensions of covariates X and matrix W"

            # self.W_ = W
            return mo.dropout(
                x=self.activation_func(
                    mo.safe_sparse_dot(a=scaled_X, b=W, backend=self.backend)
                ),
                drop_prob=self.dropout,
                seed=self.seed,
            )

        # with bias term in the hidden layer
        if W is None:
            n_features_1 = n_features + 1

            if self.nodes_sim == "sobol":
                self.W_ = generate_sobol(
                    n_dims=n_features_1,
                    n_points=self.n_hidden_features,
                    seed=self.seed,
                )
            elif self.nodes_sim == "hammersley":
                self.W_ = generate_hammersley(
                    n_dims=n_features_1,
                    n_points=self.n_hidden_features,
                    seed=self.seed,
                )
            elif self.nodes_sim == "uniform":
                self.W_ = generate_uniform(
                    n_dims=n_features_1,
                    n_points=self.n_hidden_features,
                    seed=self.seed,
                )
            else:
                self.W_ = generate_halton(
                    n_dims=n_features_1,
                    n_points=self.n_hidden_features,
                    seed=self.seed,
                )

            # self.W_ = hash_sim[self.nodes_sim](
            #         n_dims=n_features_1,
            #         n_points=self.n_hidden_features,
            #         seed=self.seed,
            #     )

            return mo.dropout(
                x=self.activation_func(
                    mo.safe_sparse_dot(
                        a=mo.cbind(
                            np.ones(scaled_X.shape[0]),
                            scaled_X,
                            backend=self.backend,
                        ),
                        b=self.W_,
                        backend=self.backend,
                    )
                ),
                drop_prob=self.dropout,
                seed=self.seed,
            )

        # W is not None
        # self.W_ = W
        return mo.dropout(
            x=self.activation_func(
                mo.safe_sparse_dot(
                    a=mo.cbind(
                        np.ones(scaled_X.shape[0]),
                        scaled_X,
                        backend=self.backend,
                    ),
                    b=W,
                    backend=self.backend,
                )
            ),
            drop_prob=self.dropout,
            seed=self.seed,
        )

    def cook_training_set(self, y=None, X=None, W=None, **kwargs):
        """Create new hidden features for training set, with hidden layer, center the response.

        Parameters:

            y: array-like, shape = [n_samples]
                Target values

            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number
                of samples and n_features is the number of features

            W: {array-like}, shape = [n_features, hidden_features]
                if provided, constructs the hidden layer via W

        Returns:

            (centered response, direct link + hidden layer matrix): {tuple}

        """

        # either X and y are stored or not
        # assert ((y is None) & (X is None)) | ((y is not None) & (X is not None))
        if self.n_hidden_features > 0:  # has a hidden layer
            assert (
                len(self.type_scaling) >= 2
            ), "must have len(self.type_scaling) >= 2 when self.n_hidden_features > 0"

        if X is None:
            if self.col_sample == 1:
                input_X = self.X_
            else:
                n_features = self.X_.shape[1]
                new_n_features = int(np.ceil(n_features * self.col_sample))
                assert (
                    new_n_features >= 1
                ), "check class attribute 'col_sample' and the number of covariates provided for X"
                np.random.seed(self.seed)
                index_col = np.random.choice(
                    range(n_features), size=new_n_features, replace=False
                )
                self.index_col_ = index_col
                input_X = self.X_[:, self.index_col_]

        else:  # X is not None # keep X vs self.X_
            if isinstance(X, pd.DataFrame):
                X = copy.deepcopy(X.values.astype(float))

            if self.col_sample == 1:
                input_X = X
            else:
                n_features = X.shape[1]
                new_n_features = int(np.ceil(n_features * self.col_sample))
                assert (
                    new_n_features >= 1
                ), "check class attribute 'col_sample' and the number of covariates provided for X"
                np.random.seed(self.seed)
                index_col = np.random.choice(
                    range(n_features), size=new_n_features, replace=False
                )
                self.index_col_ = index_col
                input_X = X[:, self.index_col_]

        if (
            self.n_clusters <= 0
        ):  # data without any clustering: self.n_clusters is None -----
            if self.n_hidden_features > 0:  # with hidden layer
                self.nn_scaler_, scaled_X = mo.scale_covariates(
                    input_X, choice=self.type_scaling[1]
                )
                Phi_X = (
                    self.create_layer(scaled_X)
                    if W is None
                    else self.create_layer(scaled_X, W=W)
                )
                Z = (
                    mo.cbind(input_X, Phi_X, backend=self.backend)
                    if self.direct_link is True
                    else Phi_X
                )
                self.scaler_, scaled_Z = mo.scale_covariates(
                    Z, choice=self.type_scaling[0]
                )
            else:  # no hidden layer
                Z = input_X
                self.scaler_, scaled_Z = mo.scale_covariates(
                    Z, choice=self.type_scaling[0]
                )
        else:  # data with clustering: self.n_clusters is not None ----- # keep
            augmented_X = mo.cbind(
                input_X,
                self.encode_clusters(input_X, **kwargs),
                backend=self.backend,
            )

            if self.n_hidden_features > 0:  # with hidden layer
                self.nn_scaler_, scaled_X = mo.scale_covariates(
                    augmented_X, choice=self.type_scaling[1]
                )
                Phi_X = (
                    self.create_layer(scaled_X)
                    if W is None
                    else self.create_layer(scaled_X, W=W)
                )
                Z = (
                    mo.cbind(augmented_X, Phi_X, backend=self.backend)
                    if self.direct_link is True
                    else Phi_X
                )
                self.scaler_, scaled_Z = mo.scale_covariates(
                    Z, choice=self.type_scaling[0]
                )
            else:  # no hidden layer
                Z = augmented_X
                self.scaler_, scaled_Z = mo.scale_covariates(
                    Z, choice=self.type_scaling[0]
                )

        # Returning model inputs -----        
        if mx.is_factor(y) is False:  # regression
            # center y
            if y is None:
                self.y_mean_, centered_y = mo.center_response(self.y_)
            else:
                self.y_mean_, centered_y = mo.center_response(y)

            # y is subsampled
            if self.row_sample < 1:
                n, p = Z.shape

                self.subsampler_ = (
                    SubSampler(
                        y=self.y_, row_sample=self.row_sample, seed=self.seed
                    )
                    if y is None
                    else SubSampler(
                        y=y, row_sample=self.row_sample, seed=self.seed
                    )
                )

                self.index_row_ = self.subsampler_.subsample()

                n_row_sample = len(self.index_row_)
                # regression
                return (
                    centered_y[self.index_row_].reshape(n_row_sample),
                    self.scaler_.transform(
                        Z[self.index_row_, :].reshape(n_row_sample, p)
                    ),
                )
            # y is not subsampled
            # regression
            return (centered_y, self.scaler_.transform(Z))

        # classification
        # y is subsampled
        if self.row_sample < 1:
            n, p = Z.shape

            self.subsampler_ = (
                SubSampler(
                    y=self.y_, row_sample=self.row_sample, seed=self.seed
                )
                if y is None
                else SubSampler(y=y, row_sample=self.row_sample, seed=self.seed)
            )

            self.index_row_ = self.subsampler_.subsample()

            n_row_sample = len(self.index_row_)
            # classification
            return (
                y[self.index_row_].reshape(n_row_sample),
                self.scaler_.transform(
                    Z[self.index_row_, :].reshape(n_row_sample, p)
                ),
            )
        # y is not subsampled
        # classification
        return (y, self.scaler_.transform(Z))

    def cook_test_set(self, X, **kwargs):
        """Transform data from test set, with hidden layer.

        Parameters:

            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number
                of samples and n_features is the number of features

            **kwargs: additional parameters to be passed to self.encode_cluster

        Returns:

            Transformed test set : {array-like}
        """

        if isinstance(X, pd.DataFrame):
            X = copy.deepcopy(X.values.astype(float))

        if (
            self.n_clusters == 0
        ):  # data without clustering: self.n_clusters is None -----
            if self.n_hidden_features > 0:
                # if hidden layer
                scaled_X = (
                    self.nn_scaler_.transform(X)
                    if (self.col_sample == 1)
                    else self.nn_scaler_.transform(X[:, self.index_col_])
                )
                Phi_X = self.create_layer(scaled_X, self.W_)
                if self.direct_link == True:
                    return self.scaler_.transform(
                        mo.cbind(scaled_X, Phi_X, backend=self.backend)
                    )
                # when self.direct_link == False
                return self.scaler_.transform(Phi_X)
            # if no hidden layer # self.n_hidden_features == 0
            return self.scaler_.transform(X)

        # data with clustering: self.n_clusters > 0 -----
        if self.col_sample == 1:
            predicted_clusters = self.encode_clusters(
                X=X, predict=True, **kwargs
            )
            augmented_X = mo.cbind(X, predicted_clusters, backend=self.backend)
        else:
            predicted_clusters = self.encode_clusters(
                X=X[:, self.index_col_], predict=True, **kwargs
            )
            augmented_X = mo.cbind(
                X[:, self.index_col_], predicted_clusters, backend=self.backend
            )

        if self.n_hidden_features > 0:  # if hidden layer
            scaled_X = self.nn_scaler_.transform(augmented_X)
            Phi_X = self.create_layer(scaled_X, self.W_)
            if self.direct_link == True:
                return self.scaler_.transform(
                    mo.cbind(augmented_X, Phi_X, backend=self.backend)
                )
            return self.scaler_.transform(Phi_X)

        # if no hidden layer
        return self.scaler_.transform(augmented_X)

    def score(self, X, y, scoring=None, **kwargs):
        """Score the model on test set features X and response y.

        Parameters:

            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number
                of samples and n_features is the number of features

            y: array-like, shape = [n_samples]
                Target values

            scoring: str
                must be in ('explained_variance', 'neg_mean_absolute_error',
                            'neg_mean_squared_error', 'neg_mean_squared_log_error',
                            'neg_median_absolute_error', 'r2')

            **kwargs: additional parameters to be passed to scoring functions

        Returns:

        model scores: {array-like}

        """

        preds = self.predict(X)

        if self.type_fit == "classification":

            if scoring is None:
                scoring = "accuracy"

            # check inputs
            assert scoring in (
                "accuracy",
                "average_precision",
                "brier_score_loss",
                "f1",
                "f1_micro",
                "f1_macro",
                "f1_weighted",
                "f1_samples",
                "neg_log_loss",
                "precision",
                "recall",
                "roc_auc",
            ), "'scoring' should be in ('accuracy', 'average_precision', \
                            'brier_score_loss', 'f1', 'f1_micro', \
                            'f1_macro', 'f1_weighted',  'f1_samples', \
                            'neg_log_loss', 'precision', 'recall', \
                            'roc_auc')"

            scoring_options = {
                "accuracy": skm.accuracy_score,
                "average_precision": skm.average_precision_score,
                "brier_score_loss": skm.brier_score_loss,
                "f1": skm.f1_score,
                "f1_micro": skm.f1_score,
                "f1_macro": skm.f1_score,
                "f1_weighted": skm.f1_score,
                "f1_samples": skm.f1_score,
                "neg_log_loss": skm.log_loss,
                "precision": skm.precision_score,
                "recall": skm.recall_score,
                "roc_auc": skm.roc_auc_score,
            }

            return scoring_options[scoring](y, preds, **kwargs)

        if self.type_fit == "regression":

            if (
                type(preds) == tuple
            ):  # if there are std. devs in the predictions
                preds = preds[0]

            if scoring is None:
                scoring = "neg_root_mean_squared_error"

            # check inputs
            assert scoring in (
                "explained_variance",
                "neg_mean_absolute_error",
                "neg_mean_squared_error",
                "neg_mean_squared_log_error",
                "neg_median_absolute_error",
                "neg_root_mean_squared_error",
                "r2",
            ), "'scoring' should be in ('explained_variance', 'neg_mean_absolute_error', \
                            'neg_mean_squared_error', 'neg_mean_squared_log_error', \
                            'neg_median_absolute_error', 'r2', 'neg_root_mean_squared_error')"

            scoring_options = {
                "neg_root_mean_squared_error": skm.root_mean_squared_error,
                "explained_variance": skm.explained_variance_score,
                "neg_mean_absolute_error": skm.median_absolute_error,
                "neg_mean_squared_error": skm.mean_squared_error,
                "neg_mean_squared_log_error": skm.mean_squared_log_error,
                "neg_median_absolute_error": skm.median_absolute_error,
                "r2": skm.r2_score,
            }

            return scoring_options[scoring](y, preds, **kwargs)



"""Model selection"""

# Authors: Thierry Moudiki
#
# License: BSD 3

# MTS -----

import numpy as np
import pandas as pd
from sklearn.model_selection import TimeSeriesSplit
from scipy.stats import describe
from tqdm import tqdm
from ..utils.timeseries import coverage, mean_errors, winkler_score


class TimeSeriesSplit(TimeSeriesSplit):
    """Time Series cross-validator"""

    def __init__(self, n_splits=5, max_train_size=None):
        super().__init__(n_splits=n_splits, max_train_size=max_train_size)

    def split(
        self,
        X,
        y=None,
        groups=None,
        initial_window=5,
        horizon=3,
        fixed_window=False,
    ):
        """Generate indices to split data into training and test set.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training data, where n_samples is the number of samples
            and n_features is the number of features.

        initial_window : int, initial number of consecutive values in each
                         training set sample

        horizon : int, number of consecutive values in test set sample

        fixed_window : boolean, if False, all training samples start at index 0

        y : array-like, shape (n_samples,)
            Always ignored, exists for compatibility.

        groups : array-like, with shape (n_samples,)
            Always ignored, exists for compatibility.

        Yields
        ------
        train : ndarray
            The training set indices for that split.

        test : ndarray
            The testing set indices for that split.
        """
        # assert initial_window
        # assert horizon
        # assert fixed_window

        try:
            n = X.shape[0]
        except:
            n = len(X)

        # Initialization of indices -----

        indices = np.arange(n)
        n_splits = 0

        # train index
        min_index_train = 0
        max_index_train = initial_window

        # test index
        min_index_test = max_index_train
        max_index_test = initial_window + horizon

        # Main loop -----

        if fixed_window == True:
            while max_index_test <= n:
                yield (
                    indices[min_index_train:max_index_train],
                    indices[min_index_test:max_index_test],
                )

                min_index_train += 1
                min_index_test += 1
                max_index_train += 1
                max_index_test += 1

                n_splits += 1

        else:
            while max_index_test <= n:
                yield (
                    indices[min_index_train:max_index_train],
                    indices[min_index_test:max_index_test],
                )

                max_index_train += 1
                min_index_test += 1
                max_index_test += 1

                n_splits += 1

        # set n_splits after (?)
        self.n_splits = n_splits
        self.max_train_size = max_index_train + 1


def cross_val_score(
    estimator,
    X,
    scoring="root_mean_squared_error",
    n_jobs=None,
    verbose=0,
    xreg=None,
    initial_window=5,
    horizon=3,
    fixed_window=False,
    show_progress=True,
    level=95,
    **kwargs,
):
    """Evaluate a score by time series cross-validation.

    Parameters:

        estimator: estimator object implementing `fit`, and of class `MTS`
            The object to use to fit the data.

        X: {array-like, sparse matrix} of shape (n_samples, n_features)
            The data to fit.

        scoring: str or a function
            A str in ('root_mean_squared_error', 'mean_squared_error', 'mean_error',
            'mean_absolute_error', 'mean_error', 'mean_percentage_error',
            'mean_absolute_percentage_error',  'winkler_score', 'coverage')
            Or a function defined as 'coverage' and 'winkler_score' in `utils.timeseries`

        n_jobs: int, default=None
            Number of jobs to run in parallel.

        verbose: int, default=0
            The verbosity level.

        xreg: array-like, optional (default=None)
            Additional (external) regressors to be passed to estimator.obj
            xreg must be in 'increasing' order (most recent observations last)

        initial_window: int
            initial number of consecutive values in each training set sample

        horizon: int
            number of consecutive values in test set sample

        fixed_window: boolean
            if False, all training samples start at index 0, and the training
            window's size is increasing.
            if True, the training window's size is fixed, and the window is
            rolling forward

        show_progress: boolean
            if True, a progress bar is printed

        **kwargs: dict
            additional parameters to be passed to `estimator`'s `fit` and `predict`

    Returns:

        A tuple: descriptive statistics or errors and raw errors

    """
    tscv = TimeSeriesSplit()

    tscv_obj = tscv.split(
        X,
        initial_window=initial_window,
        horizon=horizon,
        fixed_window=fixed_window,
    )

    if isinstance(scoring, str):

        assert scoring in (
            "root_mean_squared_error",
            "mean_squared_error",
            "mean_error",
            "mean_absolute_error",
            "mean_percentage_error",
            "mean_absolute_percentage_error",
            "winkler_score",
            "coverage",
        ), "must have scoring in ('root_mean_squared_error', 'mean_squared_error', 'mean_error', 'mean_absolute_error', 'mean_error', 'mean_percentage_error', 'mean_absolute_percentage_error',  'winkler_score', 'coverage')"

        def err_func(X_test, X_pred, scoring):
            if (estimator.replications is not None) or (
                estimator.type_pi == "gaussian"
            ):  # probabilistic
                if scoring == "winkler_score":
                    return winkler_score(X_pred, X_test, level=level)
                elif scoring == "coverage":
                    return coverage(X_pred, X_test, level=level)
                else:
                    return mean_errors(
                        pred=X_pred.mean, actual=X_test, scoring=scoring
                    )
            else:  # not probabilistic
                return mean_errors(pred=X_pred, actual=X_test, scoring=scoring)

    else:  # isinstance(scoring, str) = False

        err_func = scoring

    errors = []

    if show_progress is True:
        iterator = tqdm(tscv_obj, total=tscv.n_splits)
    else:
        iterator = tscv_obj

    for train_index, test_index in iterator:

        if verbose == 1:
            print(f"TRAIN: {train_index}")
            print(f"TEST: {test_index}")

        if isinstance(X, pd.DataFrame):
            estimator.fit(X.iloc[train_index, :], xreg=xreg, **kwargs)
            X_test = X.iloc[test_index, :]
        else:
            estimator.fit(X[train_index, :], xreg=xreg, **kwargs)
            X_test = X[test_index, :]
        X_pred = estimator.predict(
            h=int(len(test_index)), level=level, **kwargs
        )

        errors.append(err_func(X_test, X_pred, scoring))

    res = np.asarray(errors)

    return res, describe(res)



import numpy as np
from .helpers import dosubsample


class SubSampler:
    """Subsampling class.

    Attributes:

       y: array-like, shape = [n_samples]
           Target values.

       row_sample: double
           subsampling fraction

       n_samples: int
            subsampling by using the number of rows (supersedes row_sample)

       seed: int
           reproductibility seed

       n_jobs: int
            number of jobs to run in parallel

       verbose: bool
            print progress messages and bars
    """

    def __init__(
        self,
        y,
        row_sample=0.8,
        n_samples=None,
        seed=123,
        n_jobs=None,
        verbose=False,
    ):
        self.y = y
        self.n_samples = n_samples
        if self.n_samples is None:
            assert (
                row_sample < 1 and row_sample >= 0
            ), "'row_sample' must be provided, plus < 1 and >= 0"
            self.row_sample = row_sample
        else:
            assert self.n_samples < len(y), "'n_samples' must be < len(y)"
            self.row_sample = self.n_samples / len(y)
        self.seed = seed
        self.indices = None
        self.n_jobs = n_jobs
        self.verbose = verbose

    def subsample(self):
        """Returns indices of subsampled input data.

        Examples:

        <ul>
            <li> <a href="https://github.com/Techtonique/nnetsauce/blob/master/nnetsauce/demo/thierrymoudiki_20240105_subsampling.ipynb">20240105_subsampling.ipynb</a> </li>
            <li> <a href="https://github.com/Techtonique/nnetsauce/blob/master/nnetsauce/demo/thierrymoudiki_20240131_subsampling_nsamples.ipynb">20240131_subsampling_nsamples.ipynb</a> </li>
        </ul>

        """
        self.indices = dosubsample(
            y=self.y,
            row_sample=self.row_sample,
            seed=self.seed,
            n_jobs=self.n_jobs,
            verbose=self.verbose,
        )
        return self.indices



# Authors: Thierry Moudiki
#
# License: BSD 3

import numpy as np
import sklearn.metrics as skm
from .bag import RandomBag
from ..custom import CustomRegressor
from ..utils import misc as mx
from sklearn.base import RegressorMixin
import pickle
from joblib import Parallel, delayed
from tqdm import tqdm
from .helpers import rbagloop_regression


class RandomBagRegressor(RandomBag, RegressorMixin):
    """Randomized 'Bagging' Regression model

    Parameters:

        obj: object
            any object containing a method fit (obj.fit()) and a method predict
            (obj.predict())

        n_estimators: int
            number of boosting iterations

        n_hidden_features: int
            number of nodes in the hidden layer

        activation_name: str
            activation function: 'relu', 'tanh', 'sigmoid', 'prelu' or 'elu'

        a: float
            hyperparameter for 'prelu' or 'elu' activation function

        nodes_sim: str
            type of simulation for the nodes: 'sobol', 'hammersley', 'halton',
            'uniform'

        bias: boolean
            indicates if the hidden layer contains a bias term (True) or not
            (False)

        dropout: float
            regularization parameter; (random) percentage of nodes dropped out
            of the training

        direct_link: boolean
            indicates if the original predictors are included (True) in model''s
            fitting or not (False)

        n_clusters: int
            number of clusters for 'kmeans' or 'gmm' clustering (could be 0:
                no clustering)

        cluster_encode: bool
            defines how the variable containing clusters is treated (default is one-hot)
            if `False`, then labels are used, without one-hot encoding

        type_clust: str
            type of clustering method: currently k-means ('kmeans') or Gaussian
            Mixture Model ('gmm')

        type_scaling: a tuple of 3 strings
            scaling methods for inputs, hidden layer, and clustering respectively
            (and when relevant).
            Currently available: standardization ('std') or MinMax scaling ('minmax')

        col_sample: float
            percentage of covariates randomly chosen for training

        row_sample: float
            percentage of rows chosen for training, by stratified bootstrapping

        seed: int
            reproducibility seed for nodes_sim=='uniform'

        backend: str
            "cpu" or "gpu" or "tpu"

    Attributes:

        voter_: dict
            dictionary containing all the fitted base-learners


    Examples:

    ```python
    import numpy as np
    import nnetsauce as ns
    from sklearn.datasets import fetch_california_housing
    from sklearn.tree import DecisionTreeRegressor
    from sklearn.model_selection import train_test_split

    X, y = fetch_california_housing(return_X_y=True, as_frame=False)

    # split data into training test and test set
    X_train, X_test, y_train, y_test = train_test_split(X, y,
                                                        test_size=0.2, random_state=13)

    # Requires further tuning
    obj = DecisionTreeRegressor(max_depth=3, random_state=123)
    obj2 = ns.RandomBagRegressor(obj=obj, direct_link=False,
                                n_estimators=50,
                                col_sample=0.9, row_sample=0.9,
                                dropout=0, n_clusters=0, verbose=1)

    obj2.fit(X_train, y_train)

    print(np.sqrt(obj2.score(X_test, y_test))) # RMSE

    ```

    """

    # construct the object -----

    def __init__(
        self,
        obj,
        n_estimators=10,
        n_hidden_features=1,
        activation_name="relu",
        a=0.01,
        nodes_sim="sobol",
        bias=True,
        dropout=0,
        direct_link=False,
        n_clusters=2,
        cluster_encode=True,
        type_clust="kmeans",
        type_scaling=("std", "std", "std"),
        col_sample=1,
        row_sample=1,
        n_jobs=None,
        seed=123,
        verbose=1,
        backend="cpu",
    ):
        super().__init__(
            obj=obj,
            n_estimators=n_estimators,
            n_hidden_features=n_hidden_features,
            activation_name=activation_name,
            a=a,
            nodes_sim=nodes_sim,
            bias=bias,
            dropout=dropout,
            direct_link=direct_link,
            n_clusters=n_clusters,
            cluster_encode=cluster_encode,
            type_clust=type_clust,
            type_scaling=type_scaling,
            col_sample=col_sample,
            row_sample=row_sample,
            seed=seed,
            backend=backend,
        )

        self.type_fit = "regression"
        self.verbose = verbose
        self.n_jobs = n_jobs
        self.voter_ = {}

    def fit(self, X, y, **kwargs):
        """Fit Random 'Bagging' model to training data (X, y).

        Args:

            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number
                of samples and n_features is the number of features.

            y: array-like, shape = [n_samples]
                Target values.

            **kwargs: additional parameters to be passed to
                    self.cook_training_set or self.obj.fit

        Returns:

            self: object

        """

        base_learner = CustomRegressor(
            self.obj,
            n_hidden_features=self.n_hidden_features,
            activation_name=self.activation_name,
            a=self.a,
            nodes_sim=self.nodes_sim,
            bias=self.bias,
            dropout=self.dropout,
            direct_link=self.direct_link,
            n_clusters=self.n_clusters,
            type_clust=self.type_clust,
            type_scaling=self.type_scaling,
            col_sample=self.col_sample,
            row_sample=self.row_sample,
            seed=self.seed,
        )

        # 1 - Sequential training -----

        if self.n_jobs is None:
            self.voter_ = rbagloop_regression(
                base_learner, X, y, self.n_estimators, self.verbose, self.seed
            )

            self.n_estimators = len(self.voter_)

            return self

        # 2 - Parallel training -----
        # buggy
        # if self.n_jobs is not None:
        def fit_estimators(m):
            base_learner__ = pickle.loads(pickle.dumps(base_learner, -1))
            base_learner__.set_params(seed=self.seed + m * 1000)
            base_learner__.fit(X, y, **kwargs)
            return base_learner__

        if self.verbose == 1:
            voters_list = Parallel(n_jobs=self.n_jobs, prefer="threads")(
                delayed(fit_estimators)(m)
                for m in tqdm(range(self.n_estimators))
            )
        else:
            voters_list = Parallel(n_jobs=self.n_jobs, prefer="threads")(
                delayed(fit_estimators)(m) for m in range(self.n_estimators)
            )

        self.voter_ = {i: elt for i, elt in enumerate(voters_list)}

        self.n_estimators = len(self.voter_)

        return self

    def predict(self, X, weights=None, **kwargs):
        """Predict for test data X.

        Args:

            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number
                of samples and n_features is the number of features.

            **kwargs: additional parameters to be passed to
                    self.cook_test_set

        Returns:

            estimates for test data: {array-like}

        """

        def calculate_preds(voter, weights=None):
            ensemble_preds = 0

            n_iter = len(voter)

            assert n_iter > 0, "no estimator found in `RandomBag` ensemble"

            if weights is None:
                for idx, elt in voter.items():
                    ensemble_preds += elt.predict(X)

                return ensemble_preds / n_iter

            # if weights is not None:
            for idx, elt in voter.items():
                ensemble_preds += weights[idx] * elt.predict(X)

            return ensemble_preds

        # end calculate_preds ----

        if weights is None:
            return calculate_preds(self.voter_)

        # if weights is not None:
        self.weights = weights

        return calculate_preds(self.voter_, weights)



from .nodesimulation import (
    generate_sobol,
    generate_uniform,
    generate_hammersley,
    generate_halton,
)

from .getsims import getsims

__all__ = [
    "generate_sobol",
    "generate_uniform",
    "generate_hammersley",
    "generate_halton",
    "getsims",
]
